@@ -31,11 +31,12 @@ import compiler
 from compiler.ast import Import, From
 import jinja2
 import os, os.path
-import re
-import sys
 
 class InvalidTemplate(Exception):
-    """Indicates that a template has more than one separator."""
+    """Indicates that a .ks template has more than one separator."""
+
+class TemplateNotFound(Exception):
+    """Indicates that a .ks template by the given name does not exist."""
 
 class StopViewFunc(Exception):
     """Raised by return_response() to prevent templat rendering."""
@@ -47,6 +48,8 @@ def return_response(body):
     respond, bypassing templates (e.g. to return binary content
     from a file or database).
     """
+    if isinstance(body, basestring):
+        body = (body, )
     raise StopViewFunc(body)
 
 def template_filter(func):
@@ -114,8 +117,6 @@ class RenderEngine(object):
             viewfunc=viewfunc,
             body=''.join(second))
 
-        return viewfunc, ''.join(second)
-
     def compile(self, viewcode_str, filename):
         """Compile the view code and return a code object
         and dictionary of globals needed by the code object.
@@ -150,11 +151,13 @@ class RenderEngine(object):
         and template body for the .ks template at the given
         path relative to the app_dir."""
         filename = os.path.abspath(os.path.join(self.app.app_dir, name))
+        if not os.path.isfile(filename):
+            raise TemplateNotFound('could not find template %s' % name)
+
         mtime = os.stat(filename).st_mtime
         template = self.templates.get(name)
 
         if template is None or template.mtime < mtime:
-            lastmtime = mtime
             template = self.parse(file(filename, 'rb'))
             template.mtime = mtime
             template.name = name
